//  NearbyFriend
//  Copyright 2008 Urs Hengartner <uhengart@cs.uwaterloo.ca>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  There is a copy of the GNU General Public License in the COPYING file
//  packaged with this plugin; if you cannot find it, write to the Free
//  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//  02111-1307  USA

/* system headers */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

/* libgcrypt headers */
#include <gcrypt.h>

/* nbf headers */
#include "cgs97.h"
#include "serial.h"

// Germain prime from RFC 3526
static const char* DH1536_P_S = "0x"
    "FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD1"
    "29024E088A67CC74020BBEA63B139B22514A08798E3404DD"
    "EF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245"
    "E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7ED"
    "EE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3D"
    "C2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F"
    "83655D23DCA3AD961C62F356208552BB9ED529077096966D"
    "670C354E4ABC9804F1746C08CA237327FFFFFFFFFFFFFFFF";
static const char *DH1536_G_S = "0x02";
static const int DH1536_P_LEN_BITS = 1536;
//static const int DH1536_P_LEN_BYTES = 192;

static gcry_mpi_t DH1536_P = NULL;
static gcry_mpi_t DH1536_G = NULL;
static gcry_mpi_t DH1536_Q = NULL;

/* 
 * Generate a random number smaller than max and larger than 1.
 */
static gcry_mpi_t gen_random(gcry_mpi_t max)
{
    int bits, bytes;
    gcry_mpi_t ret;
    unsigned char *buf = NULL;

    bits = gcry_mpi_get_nbits(max);
    bytes = (bits + 7) >> 3;

    do {
	buf = gcry_random_bytes_secure(bytes, GCRY_STRONG_RANDOM);
	gcry_mpi_scan(&ret, GCRYMPI_FMT_USG, buf, bytes, NULL);
	gcry_free(buf);
	gcry_mpi_clear_highbit(ret, bits);
    } while (gcry_mpi_cmp(ret, max) >= 0 || 
	     gcry_mpi_cmp_ui(ret, 1) <= 0);

    return ret;
}

/*
 * Call this once, at plugin load time.  It sets up the modulus and
 * generator MPIs.
 */
void nbf_cgs97_init(void)
{
    gcry_mpi_t tmp;
    gcry_mpi_scan(&DH1536_P, GCRYMPI_FMT_HEX, DH1536_P_S, 0, NULL);
    gcry_mpi_scan(&DH1536_G, GCRYMPI_FMT_HEX, DH1536_G_S, 0, NULL);
    tmp = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_sub_ui(tmp, DH1536_P, 1);
    DH1536_Q = gcry_mpi_new(DH1536_P_LEN_BITS - 1);
    // does gcrypt support in-place operations?
    gcry_mpi_rshift(DH1536_Q, tmp, 1);  
    gcry_mpi_release(tmp);
}

/*
 * Deallocate the contents of a CGS97_Keypair (but not the CGS97_Keypair
 * itself)
 */
void nbf_cgs97_keypair_free(CGS97_Keypair *kp)
{
    gcry_mpi_release(kp->priv);
    gcry_mpi_release(kp->pub);
    kp->priv = NULL;
    kp->pub = NULL;
}

/*
 * Generate a CGS97 keypair for a specified group.
 */
gcry_error_t nbf_cgs97_gen_keypair(unsigned int groupid, CGS97_Keypair *kp)
{
    if (groupid != DH1536_GROUP_ID) {
        /* Invalid group id */
        return gcry_error(GPG_ERR_INV_VALUE);
    }

    kp->groupid = groupid;
    kp->priv = gen_random(DH1536_Q);
    kp->pub = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_powm(kp->pub, DH1536_G, kp->priv, DH1536_P);

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * Encrypt message with CGS97.
 */
gcry_error_t nbf_cgs97_encrypt(gcry_mpi_t m, gcry_mpi_t pub,
			       CGS97_Ciphertext *ct)
{
    gcry_mpi_t r, c1, c2, rm;

    /* Check that message is in range */
    if (gcry_mpi_cmp(m, DH1536_Q) >= 0) {
        return gcry_error(GPG_ERR_INV_VALUE);
    }
    
    c1 = gcry_mpi_new(DH1536_P_LEN_BITS);
    c2 = gcry_mpi_new(DH1536_P_LEN_BITS);
    rm = gcry_mpi_new(DH1536_P_LEN_BITS);
    r = gen_random(DH1536_Q);
    gcry_mpi_powm(c1, DH1536_G, r, DH1536_P);
    gcry_mpi_add(rm, r, m);
    gcry_mpi_powm(c2, pub, rm, DH1536_P);
    gcry_mpi_release(r);
    gcry_mpi_release(rm);

    ct->c1 = c1;
    ct->c2 = c2;

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * Set *answer = 1 if ciphertext decrypts to zero.
 */
gcry_error_t nbf_cgs97_decrypts_to_zero(CGS97_Ciphertext *ct, 
					CGS97_Keypair *kp, int *answer)
{    
    gcry_mpi_t tmp;

    tmp = gcry_mpi_new(DH1536_P_LEN_BITS); 
    gcry_mpi_powm(tmp, ct->c1, kp->priv, DH1536_P);
    if (gcry_mpi_cmp(tmp, ct->c2) == 0)
	*answer = 1;
    else
	*answer = 0;
    gcry_mpi_release(tmp);

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * result = ct1 + ct2;
 */
gcry_error_t nbf_cgs97_add(CGS97_Ciphertext *ct1, CGS97_Ciphertext *ct2,
			   CGS97_Ciphertext *result)
{
    gcry_mpi_t res1, res2;

    res1 = gcry_mpi_new(DH1536_P_LEN_BITS);
    res2 = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_mulm(res1, ct1->c1, ct2->c1, DH1536_P); 
    gcry_mpi_mulm(res2, ct1->c2, ct2->c2, DH1536_P); 
    result->c1 = res1;
    result->c2 = res2;

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * result = ct1 - ct2
 */
gcry_error_t nbf_cgs97_subtract(CGS97_Ciphertext *ct1, CGS97_Ciphertext *ct2,
				CGS97_Ciphertext *result)
{
    gcry_mpi_t inv1, inv2, res1, res2;
    
    inv1 = gcry_mpi_new(DH1536_P_LEN_BITS);
    inv2 = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_invm(inv1, ct2->c1, DH1536_P);
    gcry_mpi_invm(inv2, ct2->c2, DH1536_P);

    res1 = gcry_mpi_new(DH1536_P_LEN_BITS);
    res2 = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_mulm(res1, ct1->c1, inv1, DH1536_P); 
    gcry_mpi_mulm(res2, ct1->c2, inv2, DH1536_P); 
    result->c1 = res1;
    result->c2 = res2;
    gcry_mpi_release(inv1);
    gcry_mpi_release(inv2);    

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * result = ct - 1
 */
gcry_error_t nbf_cgs97_subtractOne(CGS97_Ciphertext *ct, gcry_mpi_t pub,
				   CGS97_Ciphertext *result)
{
    CGS97_Ciphertext one;

    one.c1 = gcry_mpi_set_ui(NULL, 1);
    one.c2 = pub;   
    nbf_cgs97_subtract(ct, &one, result);
    gcry_mpi_release(one.c1);
    
    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * result = ct * k
 */
gcry_error_t nbf_cgs97_multiply(CGS97_Ciphertext *ct, unsigned long k,
				CGS97_Ciphertext *result)
{
    gcry_mpi_t res1, res2, k2;

    k2 = gcry_mpi_set_ui(NULL, k);     
    res1 = gcry_mpi_new(DH1536_P_LEN_BITS);
    res2 = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_powm(res1, ct->c1, k2, DH1536_P);
    gcry_mpi_powm(res2, ct->c2, k2, DH1536_P); 
    result->c1 = res1;
    result->c2 = res2;
    gcry_mpi_release(k2);

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 *  result =  ciphertext * random value
 */
gcry_error_t nbf_cgs97_multiply_random(CGS97_Ciphertext *ct, 
				       CGS97_Ciphertext *result)
{
    gcry_mpi_t res1, res2, r;

    r = gen_random(DH1536_P);
    res1 = gcry_mpi_new(DH1536_P_LEN_BITS);
    res2 = gcry_mpi_new(DH1536_P_LEN_BITS);
    gcry_mpi_powm(res1, ct->c1, r, DH1536_P);
    gcry_mpi_powm(res2, ct->c2, r, DH1536_P); 
    result->c1 = res1;
    result->c2 = res2;
    gcry_mpi_release(r);

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * Initialize ciphertext
 */
gcry_error_t nbf_cgs97_ciphertext_init(CGS97_Ciphertext *ct, 
				       gcry_mpi_t c1, gcry_mpi_t c2)
{
    ct->c1 = c1;
    ct->c2 = c2;

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * Deallocate the contents of a CGS97_Ciphertext (but not the 
 * CGS97_Ciphertext itself)
 */
gcry_error_t nbf_cgs97_ciphertext_free(CGS97_Ciphertext *ct)
{
    gcry_mpi_release(ct->c1);
    gcry_mpi_release(ct->c2);

    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * Return length of serialized ciphertext
 */
size_t nbf_cgs97_ciphertext_len_serialized(CGS97_Ciphertext *ct)
{
    size_t len, total = 8;  /* two integers for length of each big integer */
    const enum gcry_mpi_format format = GCRYMPI_FMT_USG;

    gcry_mpi_print(format, NULL, 0, &len, ct->c1);
    total += len;
    gcry_mpi_print(format, NULL, 0, &len, ct->c2);
    total += len;
    return total;
}

/*
 * Serialize ciphertext into (allocated) buffer, buflen must be equal
 * to length of serialized ciphertext
 */
void nbf_cgs97_ciphertext_serialize(CGS97_Ciphertext *ct, unsigned char* buf,
				    int buflen)
{    
    const enum gcry_mpi_format format = GCRYMPI_FMT_USG;
    size_t len;
    size_t lenp = buflen;
    unsigned char *bufp = buf;

    gcry_mpi_print(format, NULL, 0, &len, ct->c1);
    write_mpi(ct->c1, len, "c1");
    debug_int("c1 len", bufp-len-4);

    gcry_mpi_print(format, NULL, 0, &len, ct->c2);
    write_mpi(ct->c2, len, "c2");  
    debug_int("c2 len", bufp-len-4);

    assert(lenp == 0);
}


#ifdef TEST_CGS97

int main(int argc, char **argv)
{
    CGS97_Keypair kp;
    CGS97_Ciphertext ctsquare, ctx2, cty2, ctsquare2, ctsquaresSum, ctxu2,
	ctyv2, ctsum2, ctd, ctd1, ctd2, ct0, ct1, ct2;
    int x, y, u, v;
    unsigned long square, twoX, twoY, square2;
    gcry_mpi_t m;

    if (argc != 5) {
	fprintf(stderr, "usage: %s x1 y1 x2 y2\n", argv[0]);
	exit(1);
    }

    x = atoi(argv[1]);
    y = atoi(argv[2]);
    u = atoi(argv[3]);
    v = atoi(argv[4]);

    nbf_cgs97_init();

    nbf_cgs97_gen_keypair(DH1536_GROUP_ID, &kp);
    
    // encrypt x^2+y^2
    square = x * x + y * y;
    m = gcry_mpi_set_ui(NULL, square);
    nbf_cgs97_encrypt(m, kp.pub, &ctsquare);
    gcry_mpi_release(m);

    // encrypt 2x
    twoX = 2 * x;
    m = gcry_mpi_set_ui(NULL, twoX);
    nbf_cgs97_encrypt(m, kp.pub, &ctx2);
    gcry_mpi_release(m);

    // encrypt 2y
    twoY = 2 * y;
    m = gcry_mpi_set_ui(NULL, twoY);
    nbf_cgs97_encrypt(m, kp.pub, &cty2);
    gcry_mpi_release(m);

    // u^2+v^2
    square2 = u * u + v * v;
    m = gcry_mpi_set_ui(NULL, square2);
    nbf_cgs97_encrypt(m, kp.pub, &ctsquare2); 
    gcry_mpi_release(m);

    // x^2+y^2+u^2+v^2
    nbf_cgs97_add(&ctsquare, &ctsquare2, &ctsquaresSum);

    // 2xu
    nbf_cgs97_multiply(&ctx2, u, &ctxu2);

    // 2yv
    nbf_cgs97_multiply(&cty2, v, &ctyv2);

    // 2xu + 2yv
    nbf_cgs97_add(&ctxu2, &ctyv2, &ctsum2);

    // d
    nbf_cgs97_subtract(&ctsquaresSum, &ctsum2, &ctd);
    
    // final ciphertexts
    nbf_cgs97_multiply_random(&ctd, &ct0);
    nbf_cgs97_subtractOne(&ctd, kp.pub, &ctd1);
    nbf_cgs97_multiply_random(&ctd1, &ct1);
    nbf_cgs97_subtractOne(&ctd1, kp.pub, &ctd2);
    nbf_cgs97_multiply_random(&ctd2, &ct2);

    int answer;
    nbf_cgs97_decrypts_to_zero(&ct0, &kp, &answer);
    if (answer) {
	fprintf(stdout, "friend is in same cell\n");
	exit(0);
    }
    nbf_cgs97_decrypts_to_zero(&ct1, &kp, &answer);
    if (answer) {
	fprintf(stdout, "friend is in neighbouring cell\n");
	exit(0);
    }
    nbf_cgs97_decrypts_to_zero(&ct2, &kp, &answer);
    if (answer) {
	fprintf(stdout, "friend is in diagonally adjunct cell\n");    
	exit(0);
    }
    fprintf(stdout, "friend is not nearby\n");    
    exit(0);
}

#endif
