//  NearbyFriend
//  Copyright 2008 Urs Hengartner <uhengart@cs.uwaterloo.ca>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  There is a copy of the GNU General Public License in the COPYING file
//  packaged with this plugin; if you cannot find it, write to the Free
//  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//  02111-1307  USA

/* system headers */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* This will prevent compiler errors in some instances and is better explained in the
 * how-to documents on the wiki */
#ifndef G_GNUC_NULL_TERMINATED
# if __GNUC__ >= 4
#  define G_GNUC_NULL_TERMINATED __attribute__((__sentinel__))
# else
#  define G_GNUC_NULL_TERMINATED
# endif
#endif

/* purple headers */
#include "pidgin.h"
#include "debug.h"
#include "notify.h"

/* Skyhook headers */
#include <wpsapi.h>

/* nbf headers */
#include "location.h"
#include "config.h"
#include "proto.h"
#include "ui.h"

#define SKYHOOK_USERNAME  "student"
#define SKYHOOK_REALM     "waterloo"

static NbfLocation location;

/*
 * Convert longitude/latitude, as returned by Skyhook, to x/y, as
 * required by NearbyFriend.
 * This conversion really makes sense only for computing distance between
 * devices that are in the same geographic area, which is sufficient for
 * locating nearby devices.
 * Use Mercator Projection instead?
 * XXX: This routine might fail for devices in Greenwich...
 */
static void convert_longlat_to_xy(double latitude, double longitude,
			     int* x, int* y)
{
    double difference, max_difference, max_radius;

    // difference between latitude lines is 111133 meters, latitude is
    // between -90 and 90, make sure that y is always positive
    *y = (latitude + 90) * 111133;

    // difference between longitude lines depends on latitude, longitude
    // is between -180 and 180, make sure that x is always positive
    max_radius = 6378050;    // earth radius at equator
    max_difference = 111320; // difference between two lines at equator
    difference = M_PI * max_radius * cos(latitude / 180.0 * M_PI) / 180;
    *x = longitude * difference + 180 * max_difference;
}

gboolean nbf_sample_location(gpointer data)
{
    WPS_ReturnCode rc1, rc2;
    WPS_Location* ap_location;
    WPS_IPLocation* ip_location;
    WPS_SimpleAuthentication authentication;
    int x, y;
    double longitude, latitude;
    NbfLocationType type;
    int* locationTimeoutCode = (int*) data;
    NbfgUiPrefs prefs; 

    authentication.username = SKYHOOK_USERNAME;
    authentication.realm = SKYHOOK_REALM;

    // locate device based on access points
    rc1 = WPS_location(&authentication,
	  	       WPS_NO_STREET_ADDRESS_LOOKUP,
		       &ap_location);	
    if (rc1 == WPS_OK) {	    
	longitude = ap_location->longitude;
	latitude = ap_location->latitude;	 
	convert_longlat_to_xy(latitude, longitude, &x, &y);	 
	type = NBF_LOCATION_AP;
	purple_debug_info(NBF_PLUGIN_ID, 
			  "AP longitude: %f, latitude: %f, x: %d, y: %d\n",
			  longitude, latitude, x, y);
	WPS_free_location(ap_location);
    }
    else {
	// locate device based on IP address
	rc2 = WPS_ip_location(&authentication,
		              WPS_NO_STREET_ADDRESS_LOOKUP,
			      &ip_location);	    
	if (rc2 == WPS_OK) {
	    longitude = ip_location->longitude;
	    latitude = ip_location->latitude;	
	    convert_longlat_to_xy(latitude, longitude, &x, &y);
	    type = NBF_LOCATION_IP;
	    purple_debug_info(NBF_PLUGIN_ID,
			      "IP longitude: %f, latitude: %f, x: %d, y: %d\n",
			      longitude, latitude, x, y);
	    WPS_free_ip_location(ip_location);
	    }
	else {
	  purple_debug_info(NBF_PLUGIN_ID, "No location obtained, errors were %d/%d\n", rc1, rc2);
	  goto end;
	} 
    }

    location.x = x;
    location.y = y;
    location.latitude = latitude;
    location.longitude = longitude;
    location.timestamp = nbf_get_current_time_millis();
    location.type = type;

 end:
    nbfg_ui_get_prefs(&prefs);
    purple_timeout_remove(*locationTimeoutCode);
    *locationTimeoutCode = purple_timeout_add_seconds(prefs.samplingfrequency,
						      (GSourceFunc)nbf_sample_location,
						      locationTimeoutCode);

    return TRUE;
}

/*
 * Called initially.
 */
void nbf_location_init(void)
{
    // done by purple
    //g_thread_init(NULL);
    location.x = -1;
}

/*
 * Return current location measurement, caller needs to free it 
 */
NbfLocation* nbf_get_location(void)
{
    NbfLocation* copy;

    if (location.x == -1)
	return NULL;

    copy = malloc(sizeof(NbfLocation));
    if (copy == NULL)
	return NULL;
    
    memcpy(copy, &location, sizeof(NbfLocation));
    
    return copy;
}
