//  NearbyFriend
//  Copyright 2008 Urs Hengartner <uhengart@cs.uwaterloo.ca>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  There is a copy of the GNU General Public License in the COPYING file
//  packaged with this plugin; if you cannot find it, write to the Free
//  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//  02111-1307  USA
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif

 //#include "gtk-ui.h"
/* config.h may define PURPLE_PLUGINS; protect the definition here so that we
 * don't get complaints about redefinition when it's not necessary. */
 #ifndef PURPLE_PLUGINS
 # define PURPLE_PLUGINS
 #endif

/* system headers */
#include <string.h>

/* gtk headers */
#include <glib.h>

/* This will prevent compiler errors in some instances and is better explained in the
 * how-to documents on the wiki */
#ifndef G_GNUC_NULL_TERMINATED
# if __GNUC__ >= 4
#  define G_GNUC_NULL_TERMINATED __attribute__((__sentinel__))
# else
#  define G_GNUC_NULL_TERMINATED
# endif
#endif

/* purple headers */
#include "pidgin.h"
#include "notify.h"
#include "version.h"
#include "util.h"
#include "debug.h"
#include "core.h"
#include "debug.h"
#include "notify.h"
#include "plugin.h"
#include "version.h"
#include "gtkplugin.h"

/* nbf headers */
#include "nbf-plugin.h"
#include "config.h"
#include "userstate.h"
#include "proto.h"
#include "nearbyfriend.h"
#include "environment.h"
#include "location.h"
#include "ui.h"
#include "dialogs.h"
#include "gtk-dialog.h"

#include "prefs.h"
#include "gtk-ui.h"
 
PurplePlugin *nbf_plugin_handle = NULL;

static NbfEnvironment nbf_env;

static int locationTimeoutCode;

/* Send an IM from the given account to the given recipient.  Display an
 * error dialog if that account isn't currently logged in. */
static void nbf_plugin_send_im(PurpleAccount *account, const char *recipient,
			       const char *message)
{
    PurpleConversation *conv;
    PurpleConvIm *im_conv;   
    PurpleMessageFlags msg_flags;

    conv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_IM,
            recipient, account);

    if (conv != NULL)
	im_conv = purple_conversation_get_im_data(conv);

    if (conv == NULL || im_conv == NULL) {
	const char *protocol = purple_account_get_protocol_id(account);
        const char *accountname = purple_account_get_username(account);
        PurplePlugin *p = purple_find_prpl(protocol);
        char *msg = g_strdup_printf(("You are not currently connected to "
                "account %s (%s)."), accountname,
                (p && p->info->name) ? p->info->name : ("Unknown"));
        nbf_dialog_notify_error(accountname, protocol, recipient,
                ("Not connected"), msg, NULL);
        g_free(msg);
        return;
    }
    
    // make sure that IM does not show up in sender conversation
    msg_flags = PURPLE_MESSAGE_INVISIBLE;
    purple_conv_im_send_with_flags(im_conv, message, msg_flags);
}

/* This function is the callback for the plugin action titled "Find Nearby Friends". 
 * This is the starting point for running the nearby friend protocol*/
static void
plugin_action_nearbyFriend_cb (PurplePluginAction * action)
{
	gchar msg[256];
	NbfLocation *location;

	location = nbf_get_location();
	if (location == NULL) {
	    sprintf(msg, "Your location is unknown.");
	    purple_notify_message (nbf_plugin_handle, PURPLE_NOTIFY_MSG_ERROR, 
				   "NearbyFriend Plugin",
				   msg, NULL, NULL, NULL);
	} 
	else {
	    
	    sprintf(msg,"Based on your %s, your location is:\n\nLatitude: %f, "
		    "Longitude: %f\n(x: %d, y: %d).", 
		    (location->type == NBF_LOCATION_AP ? "nearby access points" : "IP address"),
		    location->latitude, location->longitude,
		    location->x, location->y);
	    purple_notify_message (nbf_plugin_handle, PURPLE_NOTIFY_MSG_INFO, 
				   "NearbyFriend Plugin",
				   msg, NULL, NULL, NULL);
	}
}

/* we tell libpurple in the PurplePluginInfo struct to call this function to
 * get a list of plugin actions to use for the plugin.  This function gives
 * libpurple that list of actions. */
static GList *
plugin_actions (PurplePlugin * plugin, gpointer context)
{
	GList *list = NULL;
	PurplePluginAction *action = NULL;

	/* The action gets created by specifying a name to show in the UI and a
	 * callback function to call. 
	 * We add the NearbyFriend plugin callback function here*/
	action = purple_plugin_action_new ("Where am I?", 
					   plugin_action_nearbyFriend_cb);


	/* libpurple requires a GList of plugin actions, even if there is only one
	 * action in the list.  We append the action to a GList here. */
	list = g_list_append (list, action);  

	/* Once the list is complete, we send it to libpurple. */
	return list;
}

static int display_nbf_message_cb(const char *accountname,
        const char *protocol, const char *username, const char *msg)
{
    return nbf_dialog_display_nbf_message(accountname, protocol,
            username, msg);
}

static void send_im_cb(const char *accountname, const char *protocol, 
		       const char *recipient, const char *message)
{
    PurpleAccount *account = purple_accounts_find(accountname, protocol);
    if (!account) {
        PurplePlugin *p = purple_find_prpl(protocol);
        char *msg = g_strdup_printf(("Unknown account %s (%s)."),
                accountname,
                (p && p->info->name) ? p->info->name : ("Unknown"));
        nbf_dialog_notify_error(accountname, protocol, recipient,
                ("Unknown account"), msg, NULL);
        g_free(msg);
        return;
    }
    nbf_plugin_send_im(account, recipient, message);
}

static void process_new_conversation(PurpleConversation *conv)
{   
    PurpleBuddy *buddy_node;
    buddy_node = purple_find_buddy (purple_conversation_get_account(conv), 
									purple_conversation_get_name (conv));
	if (!purple_blist_node_get_bool ((PurpleBlistNode*)buddy_node, "NBF/enabled")) {
	    return;
	}
    nbf_process_new_conversation(nbf_env.userstate, &nbf_env.ui_ops, conv);  
}

static void process_delete_conversation(PurpleConversation *conv)
{   
    PurpleBuddy *buddy_node;
    buddy_node = purple_find_buddy (purple_conversation_get_account(conv), 
									purple_conversation_get_name (conv));
	if (!purple_blist_node_get_bool ((PurpleBlistNode*)buddy_node, "NBF/enabled")) {
	    return;
	}
    nbf_process_delete_conversation(nbf_env.userstate, &nbf_env.ui_ops, conv);  
}

static gboolean process_receiving_im(PurpleAccount *account, char **sender, 
        char **message, PurpleConversation *conv, PurpleMessageFlags *flags)
{
    char *username;
    const char *accountname;
    const char *protocol;
    gboolean res;
    PurpleBuddy *buddy_node;

    if (!sender || !*sender || !message || !*message)
        return 0;
	
    buddy_node = purple_find_buddy (account, *sender);
	if (!purple_blist_node_get_bool ((PurpleBlistNode*)buddy_node, "NBF/enabled")) {
	    return 0;
	}
	
    username = strdup(purple_normalize(account, *sender));
    accountname = purple_account_get_username(account);
    protocol = purple_account_get_protocol_id(account);

    res = nbf_receive_message(nbf_env.userstate, &nbf_env.ui_ops, NULL,
			      accountname, 
			      protocol, username, *message);
    
    if (res) {
	// message was for this plugin and has been processed
	free(*message);
	*message = NULL;
	return TRUE;
    }
    else
	return FALSE;    
}

static gboolean nbf_plugin_load (PurplePlugin * handle)
{
    NbfgUiPrefs prefs; 

    void *conv_handle = purple_conversations_get_handle();
  
    purple_debug_info(NBF_PLUGIN_ID, "nearbyfriend is loaded\n");

    nbf_plugin_handle = handle;

    nbf_env.userstate = nbf_userstate_create();
    nbf_env.ui_ops.send_im = send_im_cb;
    nbf_env.ui_ops.display_nbf_message = display_nbf_message_cb;

    // make sure that this plugin is called when we receive a message,
    // set priority to default plus one so that it will be called after OTR
    purple_signal_connect_priority(conv_handle, "receiving-im-msg", 
				   nbf_plugin_handle,
				   PURPLE_CALLBACK(process_receiving_im), NULL, 
				   PURPLE_SIGNAL_PRIORITY_DEFAULT + 1);

    purple_signal_connect_priority(conv_handle, "conversation-created", 
				   nbf_plugin_handle,
				   PURPLE_CALLBACK(process_new_conversation), NULL, 
				   PURPLE_SIGNAL_PRIORITY_DEFAULT + 1);
    
    purple_signal_connect_priority(conv_handle, "deleting-conversation", 
				   nbf_plugin_handle,
				   PURPLE_CALLBACK(process_delete_conversation), NULL, 
				   PURPLE_SIGNAL_PRIORITY_DEFAULT + 1);


    nbfg_ui_get_prefs(&prefs);

    // send query at defined frequency
    nbf_env.queryTimeoutCode = purple_timeout_add_seconds(prefs.queryfrequency, 
							  (GSourceFunc)nbf_send_query, 
							  &nbf_env);

    locationTimeoutCode = purple_timeout_add_seconds(prefs.samplingfrequency,
						     (GSourceFunc)nbf_sample_location,
						     &locationTimeoutCode);

    nbf_send_msg_to_everybody(NBF_MSGTYPE_PING, nbf_env.userstate, 
			      &nbf_env.ui_ops);

    nbfg_ui_init();

    return TRUE;
}

static gboolean nbf_plugin_unload(PurplePlugin *handle)
{
    purple_timeout_remove(nbf_env.queryTimeoutCode);
    purple_timeout_remove(locationTimeoutCode);

    nbf_send_msg_to_everybody(NBF_MSGTYPE_STOP, nbf_env.userstate, 
			      &nbf_env.ui_ops);

    nbf_userstate_free(nbf_env.userstate);

    return TRUE;
}

static PidginPluginUiInfo ui_info =
{
	nbfg_gtk_ui_make_widget
};
 
static PurplePluginInfo info =
{
	PURPLE_PLUGIN_MAGIC,
	PURPLE_MAJOR_VERSION,
	PURPLE_MINOR_VERSION,
	PURPLE_PLUGIN_STANDARD,					/**< type           */
	PIDGIN_PLUGIN_TYPE,						/* ui requirement */
	0,										/**< flags          */
	NULL,									/**< dependencies   */
	PURPLE_PRIORITY_DEFAULT,				/**< priority       */

	"gtk-uhengart-nearby_friend",			/* id */
	"NearbyFriend",							/* name */
	NBF_VERSION,	            			/* version*/
	"Becoming aware of nearby friends without continuous tracking",
	/* summary */
	"Become aware of nearby friends without being tracked by a third "
	"party and without continuously broadcasting your location to "
	"your friends.",
	/* description */
	"Urs Hengartner",						
	/* author */
	"http://crysp.uwaterloo.ca/software/nearbyfriend",
	/* homepage */

	nbf_plugin_load,						/* load */
	nbf_plugin_unload,						/* unload */
	NULL,                                   /**< destroy        */

	&ui_info,                               /**< ui_info        */
	NULL,									/**< extra_info     */
	NULL,									/**< prefs_info     */
	plugin_actions,							/**< actions        */
	/* padding */
	NULL,
	NULL,
	NULL,
	NULL
};
 
static void init_plugin(PurplePlugin *plugin) {
	nbfg_ui_set_ui_ops(nbfg_gtk_ui_get_ui_ops());
    nbf_dialog_set_ui_ops(nbf_gtk_dialog_get_ui_ops());

	nbf_init(NBF_VERSION_MAJOR, NBF_VERSION_MINOR);
}
 
PURPLE_INIT_PLUGIN (nearby_friend, init_plugin, info)

