//  NearbyFriend
//  Copyright 2008 Urs Hengartner <uhengart@cs.uwaterloo.ca>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  There is a copy of the GNU General Public License in the COPYING file
//  packaged with this plugin; if you cannot find it, write to the Free
//  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//  02111-1307  USA

/* system headers */
#include <assert.h>

/* purple headers */
#include "debug.h"

/* This will prevent compiler errors in some instances and is better explained in the
 * how-to documents on the wiki */
#ifndef G_GNUC_NULL_TERMINATED
# if __GNUC__ >= 4
#  define G_GNUC_NULL_TERMINATED __attribute__((__sentinel__))
# else
#  define G_GNUC_NULL_TERMINATED
# endif
#endif

/* nbf headers */
#include "nearbyfriend.h"
#include "context.h"
#include "environment.h"
#include "location.h"
#include "config.h"
#include "ui.h"

/*
 * Called once at beginning
 */
void nbf_init()
{
    nbf_cgs97_init();
    nbf_location_init();
}

/*
 * Periodically invoked to send a query to all peers
 */
gboolean nbf_send_query(gpointer data)
{  
    CGS97_Ciphertext ctsquare, ctx2, cty2;
    CGS97_Keypair* our_cgs97_key;
    char *accountname, *protocol, *user;
    unsigned long twoX, twoY;
    char *query = NULL;
    gcry_mpi_t m;
    int have_ct = 0;
    long long x, y, square;
    NbfEnvironment* env = (NbfEnvironment*) data;
    NbfUserState* userstate = env->userstate;
    NbfMessageAppOps* ui_ops = &env->ui_ops;
    NbfConnContext* context;
    GList* ims, *im;
    PurpleConversation *conv;    
    NbfLocation *location = NULL;
    NbfgUiPrefs prefs;

    memset(&ctsquare, 0, sizeof(CGS97_Ciphertext));
    memset(&ctx2, 0, sizeof(CGS97_Ciphertext));
    memset(&cty2, 0, sizeof(CGS97_Ciphertext));

    nbfg_ui_get_prefs(&prefs);

    location = nbf_get_location();
    if (location == NULL) 
        goto end;       
  
    if (nbf_get_current_time_millis() - location->timestamp > 
	prefs.locationstale * 1000) {
	purple_debug_info(NBF_PLUGIN_ID, "Our location is too old\n");
	free(location);
	goto end;
    }

    // get all IM conversations
    ims = purple_get_ims();

    for(im = ims; im != NULL; im = im->next) {
	conv = (PurpleConversation*) im->data;
	assert (conv != NULL);

	accountname = conv->account->username;
	protocol = conv->account->protocol_id;
	user = conv->name;
	
	context = nbf_context_find(userstate, user, accountname, protocol, 
				   0, NULL);
	
	if (context == NULL) { 
	    // happens if this is run before new conversation signal is sent
	    continue;
	}

	if (context->conv_state != NBF_CONV_STATE_READY)
	    continue;

	// Are we waiting for a reponse for a previous query?
	// Don't send another query, peer might have ended conversation and
	// unloaded NearbyFriend plugin, so we don't want to flood peer.
	if (context->alice_prot_state == NBF_ALICE_STATE_WAIT_RESPONSE)	    
	    continue;

	// create ciphertext only once (and only if we actually have
	// somebody to send it to)
	if (have_ct == 0) {	   

	    our_cgs97_key = &userstate->our_cgs97_key;
	  
	    // encrypt x^2+y^2
	    x = location->x / prefs.nearbydistance;
	    y = location->y / prefs.nearbydistance;

	    square = x * x + y * y;
	    m = nbf_mpi_set_long_long(square);
	    nbf_cgs97_encrypt(m, our_cgs97_key->pub, &ctsquare);
	    gcry_mpi_release(m);
	    
	    // encrypt 2x
	    twoX = 2 * x;
	    m = gcry_mpi_set_ui(NULL, twoX);
	    nbf_cgs97_encrypt(m, our_cgs97_key->pub, &ctx2);
	    gcry_mpi_release(m);
	    
	    // encrypt 2y
	    twoY = 2 * y;
	    m = gcry_mpi_set_ui(NULL, twoY);
	    nbf_cgs97_encrypt(m, our_cgs97_key->pub, &cty2);
	    gcry_mpi_release(m);
	    
	    nbf_proto_create_query(&query, our_cgs97_key->pub, 
				   &ctsquare, &ctx2, &cty2);

	    have_ct = 1;
	}
	if (ui_ops->send_im) {
	    ui_ops->send_im(accountname, protocol, user, query);
	    context->alice_prot_state = NBF_ALICE_STATE_WAIT_RESPONSE;
	}
    }
	
    free(query);
    free(location);
    nbf_cgs97_ciphertext_free(&ctsquare);
    nbf_cgs97_ciphertext_free(&ctx2);
    nbf_cgs97_ciphertext_free(&cty2);

 end:
    // re-install handler because interval could have changed
    purple_timeout_remove(env->queryTimeoutCode);
    env->queryTimeoutCode = purple_timeout_add_seconds(prefs.queryfrequency, 
						       (GSourceFunc)nbf_send_query, 
						       env);

    return TRUE;
}

/*
 * Process a received message, send a response if necessary. If it
 * returns 0, message was not for NearbyFriend. If it returns 1,
 * *newmessage might contain a new message, which should be shown to
 * user, instead of the received message. 
 */
int nbf_receive_message(NbfUserState* us, const NbfMessageAppOps *ops,
			void *opdata, const char* accountname, 
			const char* protocol, const char* sender, 
			const char *message)
{ 
    NbfConnContext *context;
    int context_added = 0;
    NbfMessageType msg_type;
    int ret = 0;
    char *response = NULL, *pong = NULL;
    char *format1, *format2, *buffer = NULL;
    PurpleAccount *account;
    PurpleConversation *conversation;
    gcry_error_t rc;
    int answer;
    CGS97_Ciphertext ctsquare, ctx2, cty2;
    CGS97_Ciphertext ctsquare2, ctsquaresSum, ctxu2, ctyv2, ctsum2;
    CGS97_Ciphertext ct1, ct2, ct3, ctd, ctd1, ctd2;
    long long u, v; 
    long long square2;
    NbfLocation *location = NULL;
    gcry_mpi_t m, pub = NULL;
    NbfgUiPrefs prefs;

    // Find our context and state with this correspondent 
    context = nbf_context_find(us, sender, accountname,
			       protocol, 1, &context_added);    

    msg_type = nbf_proto_message_type(message);

    switch(msg_type) {

    case NBF_MSGTYPE_NOTNBF:
	ret = 0;
	break;

    case NBF_MSGTYPE_PING:
	purple_debug_info(NBF_PLUGIN_ID, "Got ping\n");
	// Always accept a ping, since other side might have unloaded and
	// re-loaded plugin without us noticing or it might simply have
	// crashed.
	// If there is a conversation, respond with pong.
	account = purple_accounts_find(accountname, protocol);
	assert(account != NULL);
	conversation = purple_find_conversation_with_account(PURPLE_CONV_TYPE_IM,
							     sender, account);
	if (conversation != NULL) {
	    purple_debug_info(NBF_PLUGIN_ID, "Sent pong\n");
	    nbf_proto_create_pong(&pong);
	    if (ops->send_im)
		ops->send_im(accountname, protocol, sender, pong);
	    free(pong);
	    context->conv_state = NBF_CONV_STATE_READY;
	    context->alice_prot_state = NBF_ALICE_STATE_READY;
	    format1 = "We are running the NearbyFriend application with %s.";
	    buffer = malloc(strlen(format1) + strlen(context->username) - 1);
	    if (buffer != NULL) {
		sprintf(buffer, format1, context->username);
		if (ops->display_nbf_message) {
		    ops->display_nbf_message(accountname,
					     protocol, sender, buffer);
		}
		free(buffer);
	    }
	}
	else {
	    context->conv_state = NBF_CONV_STATE_GOT_PING;
	    // will send pong when we open conversation on this side	    
	}
	ret = 1;
	break;

    case NBF_MSGTYPE_PONG:
	purple_debug_info(NBF_PLUGIN_ID, "Got pong\n");
	if (context->conv_state == NBF_CONV_STATE_SENT_PING) {	    
	    context->conv_state = NBF_CONV_STATE_READY;
	    context->alice_prot_state = NBF_ALICE_STATE_READY;
	    format1 = "We are running the NearbyFriend application with %s.";
	    buffer = malloc(strlen(format1) + strlen(context->username) - 1);
	    if (buffer != NULL) {
		sprintf(buffer, format1, context->username);
		if (ops->display_nbf_message) {
		    ops->display_nbf_message(accountname,
					     protocol, sender, buffer);
		}
		free(buffer); 
	    }
	}
	ret = 1;
	break;
	
    case NBF_MSGTYPE_QUERY:
	if (context->conv_state == NBF_CONV_STATE_READY) {

	    format1 = "%s has queried your location.";  
	    buffer = malloc(strlen(format1) + strlen(context->username) - 1);
	    if (buffer != NULL) {
		sprintf(buffer, format1, context->username);
		if (ops->display_nbf_message) {
		    ops->display_nbf_message(accountname,
					     protocol, sender, buffer);
		}
		free(buffer); 
	    }

	    memset(&ctsquare, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctx2, 0, sizeof(CGS97_Ciphertext));
	    memset(&cty2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctsquare2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctsquaresSum, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctxu2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctyv2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctsum2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctd, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctd1, 0, sizeof(CGS97_Ciphertext));
	    memset(&ctd2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ct1, 0, sizeof(CGS97_Ciphertext));
	    memset(&ct2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ct3, 0, sizeof(CGS97_Ciphertext));

	    rc = nbf_proto_accept_query(&pub, &ctsquare, &ctx2, &cty2, 
					context, message);
	    if (rc != gcry_error(GPG_ERR_NO_ERROR))
		goto err_query;
	    
	    location = nbf_get_location();
	    if (location == NULL) 
		goto err_query;
	    
	  	
	    nbfg_ui_get_prefs(&prefs);
	  
	    if (nbf_get_current_time_millis() - location->timestamp > 
		prefs.locationstale * 1000) {
		purple_debug_info(NBF_PLUGIN_ID, "Location is too old\n");
		free(location);
		goto err_query;
	    }
	     
	    // u^2+v^2
	    u = location->x / prefs.nearbydistance; 
	    v = location->y / prefs.nearbydistance;

	    square2 = u * u + v * v;
	    free(location);   
	    m = nbf_mpi_set_long_long(square2);
	    nbf_cgs97_encrypt(m, pub, &ctsquare2); 
	    gcry_mpi_release(m);
	    
	    // x^2+y^2+u^2+v^2
	    nbf_cgs97_add(&ctsquare, &ctsquare2, &ctsquaresSum);
	    
	    // 2xu
	    nbf_cgs97_multiply(&ctx2, u, &ctxu2);
	    
	    // 2yv
	    nbf_cgs97_multiply(&cty2, v, &ctyv2);
	    
	    // 2xu + 2yv
	    nbf_cgs97_add(&ctxu2, &ctyv2, &ctsum2);
	    
	    // d
	    nbf_cgs97_subtract(&ctsquaresSum, &ctsum2, &ctd);
	    
	    // final ciphertexts
	    nbf_cgs97_multiply_random(&ctd, &ct1);
	    nbf_cgs97_subtractOne(&ctd, pub, &ctd1);
	    nbf_cgs97_multiply_random(&ctd1, &ct2);
	    nbf_cgs97_subtractOne(&ctd1, pub, &ctd2);
	    nbf_cgs97_multiply_random(&ctd2, &ct3);

	    gcry_mpi_release(pub);

	    rc = nbf_proto_create_response(&response, &ct1, &ct2, &ct3);
	    if (rc != gcry_error(GPG_ERR_NO_ERROR))
		goto err_query;

	    if (ops->send_im) {
		ops->send_im(accountname, protocol, sender, response);
	    }

	    goto no_err_query;
	    
	err_query:	
	    rc = nbf_proto_create_error_response(&response);
	    if (rc != gcry_error(GPG_ERR_NO_ERROR))
		goto no_err_query;

	    if (ops->send_im) {
		ops->send_im(accountname, protocol, sender, response);
	    }
	    
	no_err_query:
	    free(response);
	    nbf_cgs97_ciphertext_free(&ctsquare);
	    nbf_cgs97_ciphertext_free(&ctx2);
	    nbf_cgs97_ciphertext_free(&cty2);
	    nbf_cgs97_ciphertext_free(&ctsquare2);
	    nbf_cgs97_ciphertext_free(&ctsquaresSum);
	    nbf_cgs97_ciphertext_free(&ctxu2);
	    nbf_cgs97_ciphertext_free(&ctyv2);
	    nbf_cgs97_ciphertext_free(&ctsum2);
	    nbf_cgs97_ciphertext_free(&ctd);
	    nbf_cgs97_ciphertext_free(&ctd1);
	    nbf_cgs97_ciphertext_free(&ctd2);
	    nbf_cgs97_ciphertext_free(&ct1);
	    nbf_cgs97_ciphertext_free(&ct2);
	    nbf_cgs97_ciphertext_free(&ct3);
	}
	ret = 1;
	break;
	    
    case NBF_MSGTYPE_RESPONSE:
	if (context->conv_state == NBF_CONV_STATE_READY &&
	    context->alice_prot_state == NBF_ALICE_STATE_WAIT_RESPONSE) {
	    
	    memset(&ct1, 0, sizeof(CGS97_Ciphertext));
	    memset(&ct2, 0, sizeof(CGS97_Ciphertext));
	    memset(&ct3, 0, sizeof(CGS97_Ciphertext));

	    rc = nbf_proto_accept_response(&ct1, &ct2, &ct3, 
					   us, context, message);
	    if (rc != gcry_error(GPG_ERR_NO_ERROR))
		goto err_response;
		
	    nbf_cgs97_decrypts_to_zero(&ct1, &us->our_cgs97_key, 
				       &answer);
	    if (answer == 0) 
		nbf_cgs97_decrypts_to_zero(&ct2, &us->our_cgs97_key, 
					   &answer);	
	    if (answer == 0) 
		nbf_cgs97_decrypts_to_zero(&ct3, &us->our_cgs97_key, 
					   &answer);
	    
	    format1 = "%s is nearby!";
	    format2 = "%s is not nearby.";
	    buffer = malloc(strlen(format2) + strlen(context->username) - 1);
	    if (buffer != NULL) {
		if (answer) sprintf(buffer, format1, context->username);
		else sprintf(buffer, format2, context->username);	
		if (ops->display_nbf_message) {
		    ops->display_nbf_message(accountname,
					     protocol, sender, buffer);
		}
		free(buffer);
	    }

	err_response:
	    context->alice_prot_state = NBF_ALICE_STATE_READY;

	    nbf_cgs97_ciphertext_free(&ct1);
	    nbf_cgs97_ciphertext_free(&ct2);
	    nbf_cgs97_ciphertext_free(&ct3);
	}
	ret = 1;
	break;

    case NBF_MSGTYPE_STOP:
	context->conv_state = NBF_CONV_STATE_INIT;
	purple_debug_info(NBF_PLUGIN_ID, "Got a stop message\n");
	ret = 1;
	break;

    case NBF_MSGTYPE_UNKNOWN:
	purple_debug_info(NBF_PLUGIN_ID, "Got an unknown NBF message: %d\n", 
		NBF_MSGTYPE_UNKNOWN);
	ret = 1;
	break;
    }

    return ret;
}

/*
 * Invoked whenever a new conversation is set up
 */
void nbf_process_new_conversation(NbfUserState* us, 
				  const NbfMessageAppOps *ops,
				  PurpleConversation *conv)
{
    char *ping = NULL, *pong = NULL, *accountname, *protocol, *user, *format;
    PurpleConversationType type;
    NbfConnContext* context;
    int addedp;
    char *buffer;

    type = purple_conversation_get_type(conv);
    if (type != PURPLE_CONV_TYPE_IM)
	return;

    accountname = conv->account->username;
    protocol = conv->account->protocol_id;
    user = conv->name;
    
    // Typically we'll create a new context here.
    context = nbf_context_find(us, user, accountname, protocol, 
			       1, &addedp);

    if (context->conv_state == NBF_CONV_STATE_GOT_PING) {
	
	nbf_proto_create_pong(&pong);
	
	if (ops->send_im) {
	    ops->send_im(accountname, protocol, user, pong);
	    purple_debug_info(NBF_PLUGIN_ID, "Sent pong\n");
	}
	free(pong);
	context->conv_state = NBF_CONV_STATE_READY;
	context->alice_prot_state = NBF_ALICE_STATE_READY;
	format = "We are running the NearbyFriend application with %s.";
	buffer = malloc(strlen(format) + strlen(context->username) - 1);
	if (buffer != NULL) {
	    sprintf(buffer, format, context->username);
	    if (ops->display_nbf_message) {
		ops->display_nbf_message(accountname,
					 protocol, user, buffer);
	    }
	    free(buffer);
	}
    }
    else {   
	// Send a ping message to the other side, which will result in OTR
	// being set up. 
	nbf_proto_create_ping(&ping, accountname);
	
	if (ops->send_im) {
	    ops->send_im(accountname, protocol, user, ping);
	    purple_debug_info(NBF_PLUGIN_ID, "Sent ping\n");
	}
	
	context->conv_state = NBF_CONV_STATE_SENT_PING;

	free(ping);
    }
}

/*
 * Invoked whenever a conversation is deleted
 */
void nbf_process_delete_conversation(NbfUserState* us, 
				     const NbfMessageAppOps *ops,
				     PurpleConversation *conv)
{
    char *accountname, *protocol, *user;
    PurpleConversationType type;
    NbfConnContext* context;
#if 0
    char *stop = NULL;
#endif

    type = purple_conversation_get_type(conv);
    if (type != PURPLE_CONV_TYPE_IM)
	return;

    accountname = conv->account->username;
    protocol = conv->account->protocol_id;
    user = conv->name;
    
    context = nbf_context_find(us, user, accountname, protocol, 
			       0, NULL);

    assert(context != NULL);
    
    context->conv_state = NBF_CONV_STATE_INIT;

#if 0    // fails because conversation is already torn down
    nbf_proto_create_stop(&stop);
    
    if (ops->send_im) {
	ops->send_im(accountname, protocol, user, stop);
    }
    
    free(stop);
#endif
}

/*
 * Send a ping or a stop message to all IM conversations
 */
void nbf_send_msg_to_everybody(NbfMessageType type, NbfUserState* userstate, 
			       NbfMessageAppOps* ops)
{     
    GList* ims, *im;
    PurpleConversation *conv;       
    char *accountname, *protocol, *user;
    int addedp;
    char *msg;
    NbfConnContext *context;

    // get all IM conversations
    ims = purple_get_ims();

    for(im = ims; im != NULL; im = im->next) {
	conv = (PurpleConversation*) im->data;
	assert (conv != NULL);

	accountname = conv->account->username;
	protocol = conv->account->protocol_id;
	user = conv->name;
	
	// This routine is called upon loading and unloading the 
	// NearbyFriend plugin. This will create a context for each 
	// conversation in the former case.
	context = nbf_context_find(userstate, user, accountname, protocol, 1, &addedp);
	
	msg = NULL;

	if (type == NBF_MSGTYPE_STOP &&
	    context->conv_state == NBF_CONV_STATE_READY) {
	    context->conv_state = NBF_CONV_STATE_INIT;
	    nbf_proto_create_stop(&msg);	    
	}
	else if (type == NBF_MSGTYPE_PING) {
	    context->conv_state = NBF_CONV_STATE_SENT_PING;
	    nbf_proto_create_ping(&msg, accountname);
	}
	else
	    continue;

	if (ops->send_im) {
	    ops->send_im(accountname, protocol, user, msg);
	}
	free(msg);
    }
}
