//  NearbyFriend
//  Copyright 2008 Urs Hengartner <uhengart@cs.uwaterloo.ca>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of version 2 of the GNU General Public License as
//  published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  There is a copy of the GNU General Public License in the COPYING file
//  packaged with this plugin; if you cannot find it, write to the Free
//  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
//  02111-1307  USA

/* system headers */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* purple headers */
#include "debug.h"

/* nbf headers */
#include "proto.h"
#include "b64.h"
#include "serial.h"
#include "location.h"
#include "config.h"

#define PROTO_HEADER "\x00\x01\x02"

/* Return the type of the given message. */
NbfMessageType nbf_proto_message_type(const char *message)
{
    char *nbftag;

    nbftag = strstr(message, "?NBF");

    if (!nbftag) 
	return NBF_MSGTYPE_NOTNBF;

    if (!strncmp(nbftag, "?NBF:PING", 9)) return NBF_MSGTYPE_PING;
    if (!strncmp(nbftag, "?NBF:PONG", 9)) return NBF_MSGTYPE_PONG;
    if (!strncmp(nbftag, "?NBF:QUERY", 10)) return NBF_MSGTYPE_QUERY;
    if (!strncmp(nbftag, "?NBF:RESPONSE", 13)) return NBF_MSGTYPE_RESPONSE;
    if (!strncmp(nbftag, "?NBF:STOP", 9)) return NBF_MSGTYPE_STOP;

    return NBF_MSGTYPE_UNKNOWN;
}

gcry_error_t nbf_proto_create_ping(char **ping, const char* our_name)
{
    const char* format = "?NBF:PINGv1\n<b>%s</b> would like to run the "
	"NearbyFriend application with you to learn whether the two "
	"of you are nearby.\nSee <a href="
	"\"http://crysp.uwaterloo.ca/software/nearbyfriend\">"
	"http://crysp.uwaterloo.ca/software/nearbyfriend"
	"</a> for more information.";

    // Remove one "%s", add '\0'
    char* buf = malloc(strlen(format) + strlen(our_name) - 1);
    if (buf == NULL)
	return  gcry_error(GPG_ERR_ENOMEM);
    sprintf(buf, format, our_name);
    *ping = buf;
    return gcry_error(GPG_ERR_NO_ERROR);
}

gcry_error_t nbf_proto_create_pong(char **pong)
{
    char* buf = malloc(12);
    if (buf == NULL)
	return  gcry_error(GPG_ERR_ENOMEM);
    memmove(buf, "?NBF:PONGv1", 11);
    buf[11] = '\0';
    *pong = buf;
    return gcry_error(GPG_ERR_NO_ERROR);
}

gcry_error_t nbf_proto_create_query(char** query,
				    gcry_mpi_t pub,
				    CGS97_Ciphertext* ctsquare,
				    CGS97_Ciphertext* ctx2,
				    CGS97_Ciphertext* cty2)
{
    size_t buflen;
    unsigned char *bufp = NULL, *buf = NULL;
    size_t lenp, lenp1, lenp2, lenp3, keylen;
    size_t base64len;
    char *base64buf = NULL;
    gcry_error_t err;
    const enum gcry_mpi_format format = GCRYMPI_FMT_USG;

    // serialize public key and ciphertexts
    gcry_mpi_print(format, NULL, 0, &keylen, pub);    
    lenp1 = nbf_cgs97_ciphertext_len_serialized(ctsquare);
    lenp2 = nbf_cgs97_ciphertext_len_serialized(ctx2);
    lenp3 = nbf_cgs97_ciphertext_len_serialized(cty2);
    lenp = 3 + 4 + keylen + lenp1 + lenp2 + lenp3;
    bufp = malloc(lenp);
    if (bufp == NULL) {
        err = gcry_error(GPG_ERR_ENOMEM);
        goto err;
    } 
    
    buflen = lenp;
    buf = bufp;
    memmove(bufp, PROTO_HEADER, 3); 
    debug_data("Header", bufp, 3);
    bufp += 3; lenp -= 3;
    
    write_mpi(pub, keylen, "pub key");
    debug_int("key len", bufp-4-keylen);

    nbf_cgs97_ciphertext_serialize(ctsquare, bufp, lenp1);
    bufp += lenp1;
    lenp -= lenp1;
    nbf_cgs97_ciphertext_serialize(ctx2, bufp, lenp2);
    bufp += lenp2;
    lenp -= lenp2;
    nbf_cgs97_ciphertext_serialize(cty2, bufp, lenp3);
    bufp += lenp3;
    lenp -= lenp3;
    
    assert(lenp == 0);

    base64len = ((buflen + 2) / 3) * 4;
    base64buf = malloc(10 + base64len + 1 + 1);
    if (base64buf == NULL) {
        err = gcry_error(GPG_ERR_ENOMEM);
        goto err;
    }
    memmove(base64buf, "?NBF:QUERY", 10);
    nbf_base64_encode(base64buf+10, buf, buflen);
    base64buf[10 + base64len] = '.';
    base64buf[10 + base64len + 1] = '\0';

    free(buf);
    *query = base64buf;
    return gcry_error(GPG_ERR_NO_ERROR);

 err:
    free(buf);
    *query = NULL;
    return err;
}

gcry_error_t nbf_proto_create_response(char **response,
				       CGS97_Ciphertext* ct1,
				       CGS97_Ciphertext* ct2,
				       CGS97_Ciphertext* ct3)
{     
    gcry_error_t err;
    unsigned char *bufp, *buf = NULL;
    size_t lenp1, lenp2, lenp3, lenp, buflen, base64len;
    char *base64buf = NULL;

    // serialize ciphertexts
    lenp1 = nbf_cgs97_ciphertext_len_serialized(ct1);
    lenp2 = nbf_cgs97_ciphertext_len_serialized(ct2);
    lenp3 = nbf_cgs97_ciphertext_len_serialized(ct3);
    lenp = 3 + lenp1 + lenp2 + lenp3;
    bufp = malloc(lenp);
    if (bufp == NULL) {
        err = gcry_error(GPG_ERR_ENOMEM);
        goto err;
    } 

    buflen = lenp;
    buf = bufp;
    memmove(bufp, PROTO_HEADER, 3); 
    debug_data("Header", bufp, 3);
    bufp += 3; lenp -= 3;
    
    nbf_cgs97_ciphertext_serialize(ct1, bufp, lenp1);
    bufp += lenp1;
    lenp -= lenp1;
    nbf_cgs97_ciphertext_serialize(ct2, bufp, lenp2);
    bufp += lenp2;
    lenp -= lenp2;
    nbf_cgs97_ciphertext_serialize(ct3, bufp, lenp3);
    bufp += lenp3;
    lenp -= lenp3;
    
    assert(lenp == 0);

    base64len = ((buflen + 2) / 3) * 4;
    base64buf = malloc(14 + base64len + 1 + 1);
    if (base64buf == NULL) {
        err = gcry_error(GPG_ERR_ENOMEM);
        goto err;
    }
    memmove(base64buf, "?NBF:RESPONSE1", 14);
    nbf_base64_encode(base64buf+14, buf, buflen);
    base64buf[14 + base64len] = '.';
    base64buf[14 + base64len + 1] = '\0';

    free(buf);
    *response = base64buf;
    return gcry_error(GPG_ERR_NO_ERROR);

 err:
    free(buf);
    *response = NULL;
    return err;
}

gcry_error_t nbf_proto_create_error_response(char **response)
{
    char *buf;
    gcry_error_t err;

    buf = malloc(15);
    if (buf == NULL) {
        err = gcry_error(GPG_ERR_ENOMEM);
        goto err;
    }
    memmove(buf, "?NBF:RESPONSE0", 14);
    buf[14] = '\0';

    *response = buf;
    return gcry_error(GPG_ERR_NO_ERROR);

 err:
    *response = NULL;
    return err;
}

gcry_error_t nbf_proto_create_stop(char **stop)
{
    char* buf = malloc(10);
    if (buf == NULL)
	return  gcry_error(GPG_ERR_ENOMEM);
    memmove(buf, "?NBF:STOP", 9);
    buf[9] = '\0';
    *stop = buf;
    return gcry_error(GPG_ERR_NO_ERROR);
}

/*
 * Parse query message into public key and ciphertexts 
 */
gcry_error_t nbf_proto_accept_query(gcry_mpi_t* pub,
				    CGS97_Ciphertext* ctsquare, 
				    CGS97_Ciphertext* ctx2, 
				    CGS97_Ciphertext* cty2, 
				    NbfConnContext *context, 
				    const char *msg)
{ 
    char *nbftag, *endtag;
    unsigned char *rawmsg = NULL;
    size_t msglen, rawlen, lenp;
    gcry_error_t err;
    unsigned char *bufp;
    gcry_mpi_t c1 = NULL;
    gcry_mpi_t c2 = NULL;

    nbftag = strstr(msg, "?NBF:QUERY");
    if (!nbftag) {
        goto invval;
    }

    endtag = strchr(nbftag, '.');
    if (endtag) {
	msglen = endtag-nbftag;
    } else {
	goto invval;
    }

    rawlen = ((msglen-10) / 4) * 3;   /* maximum possible */
    rawmsg = malloc(rawlen);
    if (!rawmsg && rawlen > 0) {
	err = gcry_error(GPG_ERR_ENOMEM);
	goto err;
    }
    rawlen = nbf_base64_decode(rawmsg, nbftag+10, msglen-10);  /* actual size */

    bufp = rawmsg;
    lenp = rawlen;

    require_len(3);
    if (memcmp(bufp, PROTO_HEADER, 3)) {
	/* Invalid header */
	goto invval;
    }
    bufp += 3; lenp -= 3;

    read_mpi(*pub);

    read_mpi(c1);
    read_mpi(c2);
    nbf_cgs97_ciphertext_init(ctsquare, c1, c2);
    read_mpi(c1);
    read_mpi(c2);
    nbf_cgs97_ciphertext_init(ctx2, c1, c2);
    read_mpi(c1);
    read_mpi(c2);
    nbf_cgs97_ciphertext_init(cty2, c1, c2);

    if (lenp != 0) 
	goto invval;

    free(rawmsg);
    return gcry_error(GPG_ERR_NO_ERROR);
    
 invval:
    err = gcry_error(GPG_ERR_INV_VALUE);
 err:
    free(rawmsg);
    return err;
}

/*
 * Parse response message into ciphertexts 
 */
gcry_error_t nbf_proto_accept_response(CGS97_Ciphertext* ct1,
				       CGS97_Ciphertext* ct2,
				       CGS97_Ciphertext* ct3,
				       NbfUserState *userstate,
				       NbfConnContext *context, 
				       const char *msg)
{
    char *nbftag, *endtag;
    unsigned char *rawmsg = NULL;
    size_t msglen, rawlen, lenp;
    unsigned char *bufp;
    gcry_error_t err;
    gcry_mpi_t c1 = NULL;
    gcry_mpi_t c2 = NULL;

    nbftag = strstr(msg, "?NBF:RESPONSE");
    if (!nbftag) 
	goto invval;

    // Did we get error code?
    if(*(nbftag+13) == '\0' || *(nbftag+13) != '1') {
	purple_debug_info(NBF_PLUGIN_ID, "Got error response\n");
	goto invval;
    }

    endtag = strchr(nbftag, '.');
    if (endtag) {
	msglen = endtag-nbftag;
    } else {
	goto invval;
    }

    rawlen = ((msglen-14) / 4) * 3;   /* maximum possible */
    rawmsg = malloc(rawlen);
    if (!rawmsg && rawlen > 0) {
	err = gcry_error(GPG_ERR_ENOMEM);
	goto err;
    }
    rawlen = nbf_base64_decode(rawmsg, nbftag+14, msglen-14);  /* actual size */

    bufp = rawmsg;
    lenp = rawlen;

    require_len(3);
    if (memcmp(bufp, PROTO_HEADER, 3)) 
	goto invval;

    bufp += 3; lenp -= 3;

    read_mpi(c1);
    read_mpi(c2);
    nbf_cgs97_ciphertext_init(ct1, c1, c2);
    read_mpi(c1);
    read_mpi(c2);
    nbf_cgs97_ciphertext_init(ct2, c1, c2);
    read_mpi(c1);
    read_mpi(c2);
    nbf_cgs97_ciphertext_init(ct3, c1, c2);

    /* That should be everything */
    if (lenp != 0) 
	goto invval;

    free(rawmsg);
    return gcry_error(GPG_ERR_NO_ERROR);

 invval:
    err = gcry_error(GPG_ERR_INV_VALUE);
 err:
    free(rawmsg);
    return err;
}

/*
 * Return current time in milliseconds since the Epoch
 */
int nbf_get_current_time_millis(void)
{
    struct timeval ts;

    gettimeofday(&ts, NULL);
    return ts.tv_sec * 1000 + ts.tv_usec / 1000;    
}

/*
 * Initialize big integer with a 64 bit value
 */

gcry_mpi_t nbf_mpi_set_long_long(long long value)
{
#if __WORDSIZE == 64
    assert(sizeof(long long) == sizeof(long));

    return gcry_mpi_set_ui(NULL, (long) value);
#else
    gcry_mpi_t m1, m2, m3;

    assert(sizeof(long long) == 8);

    m1 = gcry_mpi_set_ui(NULL, (unsigned long) (value>>32));
    m2 = gcry_mpi_new(64);
    gcry_mpi_mul_2exp(m2, m1, 32);
    m3 = gcry_mpi_new(64);
    gcry_mpi_add_ui(m3, m2, (unsigned long) value);
    gcry_mpi_release(m1);
    gcry_mpi_release(m2);

    return m3;
#endif
}
